"""
Tools for working with Open Ephys raw data files.
"""
from __future__ import annotations

import contextlib
import functools
import logging
import os
import pathlib
import shutil
from typing import Union

import crc32c
import hdmf_zarr
import pynwb
import rich.progress
import upath

logger = logging.getLogger(__name__)

PathLike = Union[str, bytes, os.PathLike, pathlib.Path, upath.UPath]


def from_pathlike(pathlike) -> upath.UPath:
    """
    >>> from_pathlike('s3://aind-data-bucket/experiment2_Record Node 102#probeA.png')
    S3Path('s3://aind-data-bucket/experiment2_Record Node 102#probeA.png')

    >>> from_pathlike('s3://codeocean-s3datasetsbucket-1u41qdg42ur9/4797cab2-9ea2-4747-8d15-5ba064837c1c/postprocessed/experiment1_Record Node 102#Neuropix-PXI-100.ProbeA-AP_recording1/template_metrics/params.json')
    S3Path('s3://codeocean-s3datasetsbucket-1u41qdg42ur9/4797cab2-9ea2-4747-8d15-5ba064837c1c/postprocessed/experiment1_Record Node 102#Neuropix-PXI-100.ProbeA-AP_recording1/template_metrics/params.json')
    """
    if isinstance(pathlike, upath.UPath):
        return pathlike
    path: str = os.fsdecode(pathlike)
    # UPath will do rsplit('#')[0] on path
    if "#" in (p := pathlib.Path(path)).name:
        return upath.UPath(path).with_name(p.name)
    if "#" in p.parent.as_posix():
        if p.parent.as_posix().count("#") > 1:
            raise ValueError(
                f"Path {p} contains multiple '#' in a parent dirs, which we don't have a fix for yet"
            )
        for parent in p.parents:
            if "#" in parent.name:
                new = upath.UPath(path).with_name(parent.name)
                for part in p.relative_to(parent).parts:
                    new = next(
                        new.glob(part)
                    )  # we can't create or join the problem-#, so we have to 'discover' it
                return new
    return upath.UPath(path)


def checksum(path: PathLike, show_progress_bar=True) -> str:
    path = from_pathlike(path)
    hasher = crc32c.crc32c

    def formatted(x):
        return f"{x:08X}"

    blocks_per_chunk = 4096
    multi_part_threshold_gb = 0.2
    if file_size(path) < multi_part_threshold_gb * 1024**3:
        return formatted(hasher(path.read_bytes()))
    hash = 0

    with open(path, "rb") as f, get_progress():
        progress: rich.progress.Progress = globals()["progress"]
        task = progress.add_task(
            f"Checksumming {path.name}",
            total=file_size(path),
            visible=show_progress_bar,
        )
        for chunk in iter(lambda: f.read(blocks_per_chunk), b""):
            progress.update(task, advance=blocks_per_chunk)
            hash = hasher(chunk, hash)
    progress.update(task, visible=False)
    return formatted(hash)


def get_progress() -> rich.progress.Progress | contextlib.nullcontext[None]:
    if "progress" not in globals():
        globals()["progress"] = rich.progress.Progress(
            rich.progress.TextColumn("{task.description}", justify="right"),
            rich.progress.BarColumn(),
            rich.progress.TimeRemainingColumn(),
            rich.progress.FileSizeColumn(),
            rich.progress.TotalFileSizeColumn(),
        )
        return globals()["progress"]
    else:
        return contextlib.nullcontext()


def get_copy_task(src) -> rich.progress.TaskID:
    get_progress()
    progress: rich.progress.Progress = globals()["progress"]
    if not progress.tasks:
        task = progress.add_task(
            description="[cyan]Getting file sizes",
            start=False,
        )
        progress.update(task, description="[cyan]Copying files", total=size(src))
        progress.start_task(task)
    return progress.tasks[0].id


def checksums_match(*paths: PathLike) -> bool:
    checksums = tuple(checksum(p) for p in paths)
    return all(c == checksums[0] for c in checksums)


def copy(src: PathLike, dest: PathLike, max_attempts: int = 2) -> None:
    """Copy `src` to `dest` with checksum validation.

    - copies recursively if `src` is a directory
    - if dest already exists, checksums are compared, copying is skipped if they match
    - attempts to copy up to 3 times if checksums don't match
    - replaces existing symlinks with actual files
    - creates parent dirs if needed
    """
    src, dest = from_pathlike(src), from_pathlike(dest)

    with get_progress():
        progress: rich.progress.Progress = globals()["progress"]
        task = get_copy_task(src)

        if dest.exists() and dest.is_symlink():
            dest.unlink()  # we'll replace symlink with src file

        if src.is_dir():  # copy files recursively
            for path in src.iterdir():
                copy(path, dest / path.name)
            return

        if (
            not dest.suffix
        ):  # dest is a folder, but might not exist yet so can't use `is_dir`
            dest = dest / src.name
        dest.parent.mkdir(parents=True, exist_ok=True)

        if not dest.exists():
            shutil.copy2(src, dest)
            logger.debug(f"Copied {src} to {dest}")

        for _ in range(max_attempts):
            if checksums_match(src, dest):
                break
            shutil.copy2(src, dest)
        else:
            raise OSError(
                f"Failed to copy {src} to {dest} with checksum-validation after {max_attempts} attempts"
            )
        progress.update(task, advance=size(src))
        logger.debug(f"Copy of {src} at {dest} validated with checksum")


def move(src: PathLike, dest: PathLike, **rmtree_kwargs) -> None:
    """Copy `src` to `dest` with checksum validation, then delete `src`."""
    src, dest = from_pathlike(src), from_pathlike(dest)
    copy(src, dest)
    if src.is_dir():
        shutil.rmtree(src, **rmtree_kwargs)
    else:
        src.unlink()
    logger.debug(f"Deleted {src}")


def symlink(src: PathLike, dest: PathLike) -> None:
    """Create symlink at `dest` pointing to file at `src`.

    - creates symlinks recursively if `src` is a directory
    - creates parent dirs if needed (as folders, not symlinks)
    - skips if symlink already exists and points to `src`
    - replaces existing file or symlink pointing to a different location
    """
    src, dest = from_pathlike(src), from_pathlike(dest)
    if src.is_dir():
        for path in src.iterdir():
            symlink(src, dest / path.name)
    dest.parent.mkdir(parents=True, exist_ok=True)
    if dest.is_symlink() and dest.resolve() == src.resolve():
        logger.debug(f"Symlink already exists to {src} from {dest}")
        return
    with contextlib.suppress(FileNotFoundError):
        dest.unlink()
    with contextlib.suppress(FileExistsError):
        dest.symlink_to(src)
    logger.debug(f"Created symlink to {src} from {dest}")


def size(path: PathLike) -> int:
    """Return the size of a file or directory in bytes"""
    path = from_pathlike(path)
    return dir_size(path) if path.is_dir() else file_size(path)


def size_gb(path: PathLike) -> float:
    """Return the size of a file or directory in GB"""
    return round(size(path) / 1024**3, 1)


def ctime(path: PathLike) -> float:
    path = from_pathlike(path)
    with contextlib.suppress(AttributeError):
        return path.stat().st_ctime
    with contextlib.suppress(AttributeError):
        return path.stat()["LastModified"].timestamp()
    raise RuntimeError(f"Could not get size of {path}")


def file_size(path: PathLike) -> int:
    path = from_pathlike(path)
    with contextlib.suppress(AttributeError):
        return path.stat().st_size
    with contextlib.suppress(AttributeError):
        return path.stat()["size"]
    raise RuntimeError(f"Could not get size of {path}")


def dir_size(path: PathLike) -> int:
    """Return the size of a directory in bytes"""
    path = from_pathlike(path)
    if not path.is_dir():
        raise ValueError(f"{path} is not a directory")
    dir_size = 0
    dir_size += sum(file_size(f) for f in path.rglob("*") if f.is_file())
    return dir_size


def dir_size_gb(path: PathLike) -> float:
    """Return the size of a directory in GB"""
    return round(dir_size(path) / 1024**3, 1)


def free_gb(path: PathLike) -> float:
    "Return free space at `path`, to .1 GB. Raises FileNotFoundError if `path` not accessible."
    path = from_pathlike(path)
    return round(shutil.disk_usage(path).free / 1024**3, 1)


def write_nwb(
    path: PathLike,
    nwb: pynwb.NWBFile,
) -> None:
    """Switches IO based on path suffix
    - use `.nwb` to save with `NWBHDF5IO`
    - use `.nwb.zarr` to save with `NWBZarrIO` (required for reading nwb.zarr files anyway)
    - can't currently write original hdf5 NWB to cloud directly
    """
    path = from_pathlike(path)
    if path.suffix == ".zarr":
        nwb_io_class = hdmf_zarr.NWBZarrIO
        if ".nwb" not in path.name:
            path = path.with_suffix(".nwb.zarr")
    else:
        assert path.suffix == ".nwb"
        nwb_io_class = pynwb.NWBHDF5IO
        if not path.as_uri().startswith("file"):
            raise ValueError(f"Must use a local path for {nwb_io_class!r}")

    with nwb_io_class(path=path.as_posix(), mode="w") as nwb_io:
        nwb_io.write(nwb)


_NOT_FOUND = object()


class cached_property(functools.cached_property):
    """Copy of stlib functools.cached_property minus faulty thread lock.

    Issue described here: https://github.com/python/cpython/issues/87634

    This version will make concurrent tasks across multiple instances faster, but
    each instance's cached properties will no longer be thread-safe - ie. don't
    dispatch the same instance to multiple threads without implementing your own lock.
    """

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        if self.attrname is None:
            raise TypeError(
                "Cannot use cached_property instance without calling __set_name__ on it."
            )
        try:
            cache = instance.__dict__
        except (
            AttributeError
        ):  # not all objects have __dict__ (e.g. class defines slots)
            msg = (
                f"No '__dict__' attribute on {type(instance).__name__!r} "
                f"instance to cache {self.attrname!r} property."
            )
            raise TypeError(msg) from None
        val = cache.get(self.attrname, _NOT_FOUND)
        ## these lines are removed from the original cached_property ------
        # with self.lock:
        #     # check if another thread filled cache while we awaited lock
        #     val = cache.get(self.attrname, _NOT_FOUND)
        # -----------------------------------------------------------------
        if val is _NOT_FOUND:
            val = self.func(instance)
            try:
                cache[self.attrname] = val
            except TypeError:
                msg = (
                    f"The '__dict__' attribute on {type(instance).__name__!r} instance "
                    f"does not support item assignment for caching {self.attrname!r} property."
                )
                raise TypeError(msg) from None
        return val


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



import dotenv

from npc_sessions.plots import *
from npc_sessions.sessions import *
from npc_sessions.trials import *
from npc_sessions.utils import *
from npc_sessions.widgets import PSTHWidget, session_widget

_ = dotenv.load_dotenv(
    dotenv.find_dotenv(usecwd=True)
)  # take environment variables from .env

Session = DynamicRoutingSession
"""Temp alias for backwards compatibility"""



from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.figure
import matplotlib.pyplot as plt

if TYPE_CHECKING:
    import pandas as pd
    import pynwb

    import npc_sessions

import npc_sessions.plots.plot_utils as plot_utils
import npc_sessions.utils as utils


def plot_unit_quality_metrics_per_probe(session: npc_sessions.DynamicRoutingSession):
    units: pd.DataFrame = utils.good_units(session.units)

    metrics = [
        "drift_ptp",
        "isi_violations_ratio",
        "amplitude",
        "amplitude_cutoff",
        "presence_ratio",
    ]
    probes = units["electrode_group_name"].unique()

    x_labels = {
        "presence_ratio": "fraction of session",
        "isi_violations_ratio": "violation rate",
        "drift_ptp": "microns",
        "amplitude": "uV",
        "amplitude_cutoff": "frequency",
    }

    for metric in metrics:
        fig, ax = plt.subplots(1, len(probes))
        probe_index = 0
        fig.suptitle(f"{metric}")
        for probe in probes:
            units_probe_metric = units[units["electrode_group_name"] == probe][metric]
            ax[probe_index].hist(units_probe_metric, bins=20, density=True)
            ax[probe_index].set_title(f"{probe}")
            ax[probe_index].set_xlabel(x_labels[metric])
            probe_index += 1

        fig.set_size_inches([10, 6])
    plt.tight_layout()


def plot_all_unit_spike_histograms(
    session: npc_sessions.DynamicRoutingSession,
) -> tuple[matplotlib.figure.Figure, ...]:  # -> tuple:# -> tuple:
    session.units[:].query("default_qc")
    figs: list[matplotlib.figure.Figure] = []
    for obj in session.all_spike_histograms.children:
        fig, ax = plt.subplots()
        ax.plot(obj.timestamps, obj.data, linewidth=0.1, alpha=0.8, color="k")
        plot_utils.add_epoch_color_bars(
            ax, session.epochs[:], y=50, va="bottom", rotation=90
        )
        ax.set_title(obj.description, fontsize=8)
        fig.suptitle(session.session_id, fontsize=10)
        ax.set_xlabel(obj.timestamps_unit)
        ax.set_ylabel(obj.unit)
        ax.set_ylim([0, ax.get_ylim()[1]])
        ax.margins(0)
        ax.set_frame_on(False)
        fig.set_layout_engine("tight")
        fig.set_size_inches(5, 5)
        figs.append(fig)
    return tuple(figs)


def plot_unit_spikes_channels(
    session: npc_sessions.DynamicRoutingSession,
    lower_channel: int = 0,
    upper_channel: int = 384,
):
    units: pd.DataFrame = utils.good_units(session.units)

    probes = units["electrode_group_name"].unique()
    for probe in probes:
        fig, ax = plt.subplots()
        unit_spike_times = units[units["electrode_group_name"] == probe]
        unit_spike_times_channel = unit_spike_times[
            (unit_spike_times["peak_channel"] >= lower_channel)
            & (unit_spike_times["peak_channel"] <= upper_channel)
        ]["spike_times"].to_numpy()
        hist, bins = utils.bin_spike_times(unit_spike_times_channel, bin_interval=1)

        ax.plot(hist)
        ax.set_title(
            f"{probe} spike hist for channel range {lower_channel} to {upper_channel}"
        )
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Spike Count per 1 second bin")


def plot_drift_maps(
    session: npc_sessions.DynamicRoutingSession | pynwb.NWBFile,
) -> tuple[plt.Figure, ...]:
    figs = []
    for k, v in session.analysis["drift_maps"].images.items():
        fig, ax = plt.subplots()
        ax.imshow(v)
        fig.suptitle(f"{session.session_id}")
        ax.set_title(k, fontsize=8)
        ax.margins(0)
        ax.axis("off")
        fig.set_size_inches([5, 5])
        fig.tight_layout()
        figs.append(fig)
    return tuple(figs)



from __future__ import annotations

import importlib.resources
import logging
import os
import subprocess
import sys

import npc_lims
import npc_session
import upath

import npc_sessions

logger = logging.getLogger(__name__)

QC_REPO = npc_lims.DR_DATA_REPO.parent.parent / "qc"

QC_NOTEBOOK = npc_sessions.from_pathlike(
    importlib.resources.files("npc_sessions") / "notebooks" / "dynamic_routing_qc.ipynb"
)


def write_qc_notebook(
    session: str | npc_session.SessionRecord | npc_sessions.DynamicRoutingSession,
    **session_kwargs,
) -> upath.UPath:
    """Execute and save (as .ipynb) the DR QC notebook for a given session."""
    if not isinstance(session, npc_sessions.DynamicRoutingSession):
        session = npc_sessions.DynamicRoutingSession(session, **session_kwargs)

    # pass session to run in notebook via env var
    if session.info:
        var = session.info.allen_path.as_posix()
    else:
        var = str(session.id)
    os.environ["NPC_SESSION_ID"] = var

    new_name = session.id if not session.info else session.info.allen_path.stem
    logger.info(f"running {QC_NOTEBOOK.name} for {session.id}")
    subprocess.run(
        f"jupyter nbconvert --to notebook --execute {QC_NOTEBOOK} --allow-errors --output {new_name}",
        shell=True,
        check=True,
        capture_output=False,
        env=os.environ,
    )
    return QC_NOTEBOOK.with_name(f"{new_name}.ipynb")


def move_to_s3(file: npc_sessions.PathLike) -> None:
    """copy to s3, remove local copy"""
    file = npc_sessions.from_pathlike(file)
    (s3 := QC_REPO / file.name).write_bytes(file.read_bytes())
    file.unlink()
    logger.info(f"moved {file.name} to {s3.parent}")


def main() -> None:
    npc_sessions.assert_s3_write_credentials()
    sessions = sys.argv[1:] or (s.id for s in npc_lims.get_session_info())
    for session in sessions:
        path = write_qc_notebook(session)
        move_to_s3(path)


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING

import matplotlib.figure
import matplotlib.pyplot as plt
import numpy as np

if TYPE_CHECKING:
    import pandas as pd

    import npc_sessions
import npc_sessions.plots.plot_utils as plot_utils


def plot_performance_by_block(
    session: "npc_sessions.DynamicRoutingSession",
) -> matplotlib.figure.Figure:
    task_performance_by_block_df: pd.DataFrame = session.analysis["performance"][:]

    n_passing_blocks = np.sum(task_performance_by_block_df["cross_modal_dprime"] >= 1.5)
    failed_block_ind = task_performance_by_block_df["cross_modal_dprime"] < 1.5

    # blockwise behavioral performance
    xvect = task_performance_by_block_df.index.values
    fig, ax = plt.subplots(2, 1)
    ax[0].plot(
        xvect,
        task_performance_by_block_df["signed_cross_modal_dprime"],
        "ko-",
        label="cross-modal",
    )
    ax[0].plot(
        xvect[failed_block_ind],
        task_performance_by_block_df["signed_cross_modal_dprime"][failed_block_ind],
        "ro",
        label="failed",
    )
    ax[0].axhline(0, color="k", linestyle="--", linewidth=0.5)
    ax[0].set_title(
        "cross-modal dprime: "
        + str(n_passing_blocks)
        + "/"
        + str(len(task_performance_by_block_df))
        + " blocks passed"
    )
    ax[0].set_ylabel("aud <- dprime -> vis")

    ax[1].plot(
        xvect, task_performance_by_block_df["vis_intra_dprime"], "go-", label="vis"
    )
    ax[1].plot(
        xvect, task_performance_by_block_df["aud_intra_dprime"], "bo-", label="aud"
    )
    ax[1].set_title("intra-modal dprime")
    ax[1].legend(["vis", "aud"])
    ax[1].set_xlabel("block index")
    ax[1].set_ylabel("dprime")

    fig.suptitle(session.id)
    fig.tight_layout()

    return fig


def plot_first_lick_latency_hist(
    session: "npc_sessions.DynamicRoutingSession",
) -> matplotlib.figure.Figure:
    # first lick latency histogram

    trials: pd.DataFrame = session.trials[:]

    xbins = np.arange(0, 1, 0.05)
    fig, ax = plt.subplots(1, 1)
    ax.hist(
        trials.query("is_vis_stim==True")["response_time"]
        - trials.query("is_vis_stim==True")["stim_start_time"],
        bins=xbins,
        alpha=0.5,
    )

    ax.hist(
        trials.query("is_aud_stim==True")["response_time"]
        - trials.query("is_aud_stim==True")["stim_start_time"],
        bins=xbins,
        alpha=0.5,
    )

    ax.legend(["vis stim", "aud stim"])
    ax.set_xlabel("lick latency (s)")
    ax.set_ylabel("trial count")
    ax.set_title("lick latency: " + session.id)

    return fig


def plot_lick_raster(
    session: "npc_sessions.DynamicRoutingSession",
) -> matplotlib.figure.Figure:
    timeseries = session.processing["behavior"]["licks"]
    trials: pd.DataFrame = session.trials[:]

    fig, ax = plt.subplots(1, 1)
    ax.axvline(0, color="k", linestyle="--", linewidth=0.5)
    for tt, trial in trials.iterrows():
        trial_licks = (
            timeseries.timestamps[
                (timeseries.timestamps > trial["stim_start_time"] - 1)
                & (timeseries.timestamps < trial["stim_start_time"] + 2)
            ]
            - trial["stim_start_time"]
        )

        ax.vlines(trial_licks, tt, tt + 1)

    ax.set_xlim([-1, 2])
    ax.set_xlabel("time rel to stim onset (s)")
    ax.set_ylabel("trial number")
    ax.set_title(timeseries.description, fontsize=8)
    fig.suptitle(session.id, fontsize=10)

    return fig


def plot_running(
    session: "npc_sessions.DynamicRoutingSession",
) -> matplotlib.figure.Figure:
    timeseries = session.processing["behavior"]["running_speed"]
    epochs: pd.DataFrame = session.epochs[:]
    plt.style.use("seaborn-v0_8-notebook")

    fig, ax = plt.subplots()

    for _, epoch in epochs.iterrows():
        epoch_indices = (timeseries.timestamps >= epoch["start_time"]) & (
            timeseries.timestamps <= epoch["stop_time"]
        )
        if len(epoch_indices) > 0:
            ax.plot(
                timeseries.timestamps[epoch_indices],
                timeseries.data[epoch_indices],
                linewidth=0.1,
                alpha=0.8,
                color="k",
            )
    k = 100 if "cm" in timeseries.unit else 1
    ymax = 0.8 * k
    plot_utils.add_epoch_color_bars(ax, epochs, rotation=90, y=ymax, va="top")
    ax.set_ylim([-0.05 * k, ymax])
    ax.hlines(
        0,
        0,
        max(timeseries.timestamps),
        color="k",
        linestyle="--",
        linewidth=0.5,
        zorder=0,
    )
    ax.margins(0)
    ax.set_frame_on(False)
    ax.set_ylabel(timeseries.unit)
    ax.set_xlabel(timeseries.timestamps_unit)
    title = timeseries.description
    if max(timeseries.data) > ax.get_ylim()[1]:
        title += f"\ndata clipped: {round(max(timeseries.data)) = } {timeseries.unit} at {timeseries.timestamps[np.argmax(timeseries.data)]} {timeseries.timestamps_unit}"
    ax.set_title(title, fontsize=8)
    fig.suptitle(session.id, fontsize=10)
    fig.set_size_inches(10, 4)
    fig.set_layout_engine("tight")
    return fig



import random
from typing import TYPE_CHECKING

import matplotlib.figure
import matplotlib.pyplot as plt
import numpy as np
import rich

if TYPE_CHECKING:
    import npc_sessions

import npc_sessions.utils as utils


def plot_bad_lick_times(
    session: "npc_sessions.DynamicRoutingSession",
) -> tuple[plt.Figure, ...]:
    """A loop making eventplots vsyncs for trials with:
    - licks in script but no lick within response window
    - licks not in script, but lick within response window
    """
    assert session.is_sync and session._trials._sync is not None

    trials_with_lick_outside_response_window = (
        session.trials[:]
        .query("is_response")
        .query(
            "response_window_start_time > response_time or response_window_stop_time < response_time"
        )
    ).index.tolist()

    trials_with_lick_inside_response_window_but_not_recorded = (
        session.trials[:]
        .query("not is_response")
        .query(
            "response_window_start_time <= response_time <= response_window_stop_time"
        )
    ).index

    figs = []
    for idx in (
        *trials_with_lick_outside_response_window,
        *trials_with_lick_inside_response_window_but_not_recorded,
    ):
        figs.append(plot_trial_lick_timing(session, idx))
    return tuple(figs)


def plot_assorted_lick_times(
    session: "npc_sessions.DynamicRoutingSession",
) -> tuple[plt.Figure, ...]:
    sync_time = session._trials.response_time
    script_time = utils.safe_index(
        session._trials._flip_times, session._trials._sam.trialResponseFrame
    )
    intervals = np.abs(sync_time - script_time)
    figs = []
    for idx, trial_idx in enumerate(
        [
            np.nanargmax(intervals),
            np.nanargmin(intervals),
            random.choice(session.trials[:].query("is_response").index),
        ]
    ):
        fig = session.plot_trial_lick_timing(trial_idx)  # type: ignore[attr-defined]
        fig.axes[0].set_title(
            fig.axes[0].get_title()
            + " - "
            + ["longest", "shortest", "random"][idx]
            + " lick-lickFrame interval"
        )
        figs.append(fig)
    return tuple(figs)


def plot_trial_lick_timing(
    session: "npc_sessions.DynamicRoutingSession", trial_idx: int
) -> matplotlib.figure.Figure:
    if not session.is_sync or session._trials._sync is None:
        raise ValueError("session must have sync data")
    start = session._trials.response_window_start_time[trial_idx]
    stop = session._trials.response_window_stop_time[trial_idx]
    vsyncs = session._trials._sync.get_falling_edges("vsync_stim", units="seconds")
    lick_sensor_rising = session._trials._sync.get_rising_edges(
        "lick_sensor", units="seconds"
    )
    lick_sensor_falling = session._trials._sync.get_falling_edges(
        "lick_sensor", units="seconds"
    )

    fig, ax = plt.subplots()
    padding = 0.3
    marker_config = {"linestyles": "-", "linelengths": 0.2}
    line_config = {"linestyles": "-", "linelengths": 1}
    ax.eventplot(
        vsyncs[(vsyncs >= start - padding) & (vsyncs <= stop + padding)],
        **line_config,
        label="vsyncs",
        alpha=0.5,
        color="orange",
    )
    ax.eventplot(
        vsyncs[(vsyncs >= start) & (vsyncs <= stop)],
        **line_config,
        label="vsyncs within response window",
        color="orange",
    )
    ax.eventplot(
        lick_sensor_rising[
            (lick_sensor_rising >= start - padding)
            & (lick_sensor_rising <= stop + padding)
        ],
        **marker_config,
        label="lick sensor rising",
        color="k",
        lineoffsets=1,
    )
    ax.eventplot(
        lick_sensor_falling[
            (lick_sensor_falling >= start - padding)
            & (lick_sensor_falling <= stop + padding)
        ],
        **marker_config,
        label="lick sensor falling",
        color="grey",
        lineoffsets=1,
    )
    ax.eventplot(
        [
            utils.safe_index(
                session._trials._input_data_times,
                session._trials._sam.trialResponseFrame[trial_idx],
            )
        ],
        **marker_config,
        label="responseFrame in TaskControl",
        color="lime",
        lineoffsets=1.6,
    )

    ax.eventplot(
        [
            (
                lick_frames := utils.safe_index(
                    session._trials._input_data_times, session._trials._sam.lickFrames
                )
            )[(lick_frames >= start) & (lick_frames <= stop)]
        ],
        **(marker_config | {"linestyles": ":"}),
        label="lickFrames in TaskControl",
        color="g",
        lineoffsets=1.6,
    )

    ax.eventplot(
        [
            utils.safe_index(
                session._trials._flip_times,
                session._trials._sam.stimStartFrame[trial_idx],
            )
        ],
        **marker_config,
        label="stim start frame in TaskControl",
        color="r",
        lineoffsets=1.6,
    )

    ax.legend(fontsize=8, loc="upper center", fancybox=True, ncol=4)
    ax.set_yticks([])
    ax.set_xlabel("time (s)")
    ax.title.set_text(
        f"{session.id} - sync & script timing - trial {trial_idx} - {session._trials.stim_name[trial_idx]}"
    )
    fig.set_size_inches(12, 4)
    return fig


def plot_lick_times_on_sync_and_script(
    session: "npc_sessions.DynamicRoutingSession",
) -> tuple[plt.Figure, plt.Figure]:
    """-stem plot of lick times on sync relative to lick times in TaskControl
    - histogram showing distribution of same intervals
    """
    sync_time = session._trials.response_time
    script_time = utils.safe_index(
        session._trials._flip_times, session._trials._sam.trialResponseFrame
    )

    intervals = sync_time - script_time
    fig1, ax = plt.subplots()
    markerline, stemline, baseline = plt.stem(
        sync_time,
        intervals,
        bottom=0,
        orientation="horizontal",
    )
    plt.setp(stemline, linewidth=0.5, alpha=0.3)
    plt.setp(markerline, markersize=0.5, alpha=0.8)
    plt.setp(baseline, visible=False)
    ax.set_xlabel("lick time on sync relative to lick time in TaskControl (s)")
    ax.set_ylabel("experiment time (s)")
    ax.set_title(f"{np.nanmean(intervals) = :.3f}s, {np.nanstd(intervals) = :.3f}")

    fig2, ax = plt.subplots()
    ax.hist(intervals, bins=50)
    ax.set_xlabel("lick time on sync relative to lick time in TaskControl (s)")
    ax.set_ylabel("count")
    ax.set_title(f"{np.nanmean(intervals) = :.3f}s, {np.nanstd(intervals) = :.3f}")
    return fig1, fig2


def plot_diode_flip_intervals(
    session: "npc_sessions.DynamicRoutingSession",
) -> matplotlib.figure.Figure:
    fig = session.sync_data.plot_diode_measured_sync_square_flips()
    names = tuple(
        k for k, v in session.stim_frame_times.items() if not isinstance(v, Exception)
    )
    for idx, ax in enumerate(fig.axes):
        if len(names) == len(fig.axes):
            ax.set_title(names[idx].split("_")[0])
    fig.set_size_inches(12, 6)
    return fig


def plot_vsyncs_and_diode_flips_at_ends_of_each_stim(
    session: "npc_sessions.DynamicRoutingSession",
) -> matplotlib.figure.Figure:
    rich.print("[bold] Fraction long frames [/bold]")
    for stim_name, stim_times in session.stim_frame_times.items():
        if isinstance(stim_times, Exception):
            continue
        intervals = np.diff(stim_times)
        fraction_long = np.sum(intervals > 0.02) / len(intervals)
        longest_interval = max(intervals)
        start_tag, end_tag = (
            ("[bold green]", "[/bold green]")
            if fraction_long < 0.01 and longest_interval < 0.5
            else ("[bold magenta]", "[/bold magenta]")
        )
        rich.print(
            start_tag
            + stim_name
            + ": "
            + str(fraction_long)
            + " \t\t longest interval:"
            + str(longest_interval)
            + end_tag
        )

    # TODO switch this to get epoch start/ stop times and plot only for good stimuli
    names = tuple(
        k for k, v in session.stim_frame_times.items() if not isinstance(v, Exception)
    )
    fig = session.sync_data.plot_stim_onsets()
    for idx, ax in enumerate(fig.axes):
        if len(names) == len(fig.axes):
            ax.set_title(names[idx].split("_")[0])
    fig.set_size_inches(10, 5 * len(fig.axes))
    fig.subplots_adjust(hspace=0.3)

    fig = session.sync_data.plot_stim_offsets()
    names = tuple(k for k, v in session.stim_frame_times.items() if v is not None)
    for idx, ax in enumerate(fig.axes):
        if len(names) == len(fig.axes):
            ax.set_title(names[idx].split("_")[0])
    fig.set_size_inches(10, 5 * len(fig.axes))
    fig.subplots_adjust(hspace=0.3)

    return fig


def plot_histogram_of_vsync_intervals(session):
    stim_frame_times = {
        k: v
        for k, v in session.stim_frame_times.items()
        if not isinstance(v, Exception)
    }

    fig_hist, axes_hist = plt.subplots(1, len(stim_frame_times))
    fig_hist.set_size_inches(12, 6)

    for ax, (stim_name, stim_times) in zip(axes_hist, stim_frame_times.items()):
        ax.hist(np.diff(stim_times), bins=np.arange(0, 0.1, 0.001))
        ax.set_yscale("log")
        ax.axvline(1 / 60, c="k", ls="dotted")
        ax.set_title(stim_name.split("_")[0])
        ax.set_xlabel("time (s)")
        ax.set_ylabel("frame interval count")
    plt.tight_layout()


def plot_reward_times(session):
    fig, ax = plt.subplots()
    ax.hist(session.trials[:].reward_time - session.trials[:].response_time)
    ax.xaxis.label.set_text("contingent_reward_time - response_time (s)")
    ax.yaxis.label.set_text("count")
    ax.set_xlim(min(0, ax.get_xlim()[0]), ax.get_xlim()[1])
    ax.vlines(0, 0, ax.get_ylim()[1], color="k", linestyle="dotted")

    return fig


def plot_long_vsyncs_distribution_across_trial(
    session: "npc_sessions.DynamicRoutingSession",
) -> matplotlib.figure.Figure:
    all_vsyncs = np.hstack(session.sync_data.vsync_times_in_blocks)

    interval_threshold = 0.017  # s

    all_long_intervals: list[float] = []
    fig, axes = plt.subplots(1, 2, sharey=True, sharex=True, figsize=(12, 4))
    for cidx, condition in enumerate(("is_vis_stim", "is_aud_stim")):
        plt.sca(axes[cidx])
        for idx, trial in session.trials[:].query(condition).iterrows():
            vsyncs = (
                all_vsyncs[
                    (all_vsyncs >= trial.start_time) & (all_vsyncs <= trial.stop_time)
                ]
                - trial.start_time
            )
            intervals = np.diff(vsyncs)
            long_vsyncs = vsyncs[:-1][intervals > interval_threshold]
            long_intervals = intervals[intervals > interval_threshold]
            plt.scatter(
                long_vsyncs,
                np.ones_like(long_vsyncs) * idx,
                c=long_intervals,
                marker=".",
                facecolor="none",
                cmap="magma_r",
                s=4,
                alpha=1,
            )
            all_long_intervals.extend(long_intervals)

        plt.gca().set_title(condition.split("_")[1])
        top_ax = plt.gca().secondary_xaxis("top")
        top_ax.set_xticks([trial["stim_start_time"] - trial.start_time])
        top_ax.set_xticklabels(
            [f'stim_start_time {" (incl device latency)" if session.is_ephys else ""}']
        )
        c = plt.colorbar()
        c.set_ticks(
            [
                interval_threshold,
                *(unique_intervals := np.unique(np.round(all_long_intervals, 2))),
            ]
        )
        plt.clim([interval_threshold, max(unique_intervals)])
        plt.xlabel("time from trial start (s)")
        if cidx == 0:
            plt.ylabel("trial index")

    plt.suptitle(
        f"{session.id} - vsync intervals > {interval_threshold:.3f} s", fontsize=8
    )
    return fig
